# Detector

## Installation

To install from the CCF Python Package Index:

```
$ python3 -m pip install --extra-index-url https://package.computecanada.ca/pypi druid-detector
```

This relies on there not being a package called "druid-detector" on the main
PyPi site; however, if only `--index-url` were specified, any dependencies
would be expected to come from the CCF index as well, which is not the case.

### Configuration

A sample configuration is included in `[detector.conf.sample]` (and can be
generated using the `--sample-config` command-line switch).

Run-time configuration can also be managed with the use of environment
variables and, in some cases, command-line switches.  Items in the
configuration file will be overridden by environment variables matching the
pattern `BEAM_{SECTION}_{ITEM}`.  For example, to override the Manager URI,
which is configured in the `[manager]` section as `uri`, set
`BEAM_MANAGER_URI` in the environment to the desired value.

Command-line switches, if defined for the configuration item, will follow a
pattern `--{section}-{item}`.  The Bursts detector's command-line switch for
the `many` item is therefore `--bursts-many`.  Available switches may be
viewed by invoking the Detector application with `--help`.

## Development

### Workspace setup

* Set up your workspace:
    ```
    git clone gitlab@git.computecanada.ca:frak/burst/detector.git
    cd detector

    # register submodules
    git submodule init

    # update local copy of submodules
    git submodule update
    ```
* Invoke test scriptage prior to any commits using a pre-commit hook:
    ```
    cd .git/hooks
    ln -s ../../tests/test-all pre-commit
    cd ../..
    ```
* Create virtual environment for testing:
    ```
    python3 -m venv venv

    # activate environment
    . venv/bin/activate

    # install project requirements
    pip install -r requirements.txt

    # install testing requirements
    pip install -r tests/requirements.txt
    ```

    The testing requires the local environment include the necessary Python
    modules so remember to invoke `. venv/bin/activate` in any session where
    testing is launched.

### Detector module development

To develop a new detection module, subclass the Detector class with a new
definition file under `detectors/`.  Copy an existing module, such as
`oldjobs.py`, and remove any specific code until there's a skeleton.

Give it a shortish but descriptive name so its intended target is immediately
obvious, such as ResourceWastage.  You'll also need an identifier to be used
internally.  This should be the same as the name of the class, but lowercase:
`resourcewastage`.  This is a class variable `_name` set at the top of the
class definition.  This is used in reporting detected cases and will match a
case subclass in the Manager.

#### Methods to define

The new class must override the following methods in the Detector class:

1. `define_configuration()`: **This method must start by calling the parent's
method.**  Afterwards, use the mergeconf library's `add()` method to add any
necessary configuration items to the given configuration object.  This
integrates the new module with the application's configuration mechanism.

2. `detect()`: Run whatever queries and analysis necessary for detection and
return a list of detected cases.  The expected format is described in the
Detector class's docstrings.
    This method takes no parameters, but the module's configuration is
available in `cls._config` and any configuration item can be read using the
mergeconf library's API.  (For example, if a configuration item `colour` was
created, its configured value can be retrieved with `cls._config.colour`.)

#### Module registration

To be known and used, the module must be registered.  This is done in
`detectors/__init__.py`.  Follow the pattern established for other modules.

#### Configuration file

Configuration can be enacted through files, environment variables, or via the
command line (to a limited extent, as defined by the detector).  Files are
more straightforward for development and testing.   Apart from the Manager
configuration section, each detection module must define a configuration
section.  At minimum, this will contain the following:

```
[oldjobs]
enable = true
```

This simple snippets configures the `oldjobs` detector to be enabled.  Other
configuration items defined in the detector module can be set here (and
must be if they were defined as mandatory).

Configuration files are searched for in the following locations, first found
wins:

* `detector.conf` in current working directory
* `$HOME/.config/druid/detector.conf`
* `/etc/druid/detector.conf`

#### Testing

Development and testing must be performed against a non-production instance of
the Manager; early on it might be a good idea to set the server to something
nonsensical to avoid any accidental reports.  In addition, use the `--dry-run`
command-line switch to print the intended report to the console instead of
sending it to the manager.

## Packaging and deployment

CI jobs automatically package and deploy the Detector app and modules.  All
commits result in packages uploaded to the project's [package
repository](https://git.computecanada.ca/frak/burst/detector/-/packages);
tagged releases result in the package being available via Compute Canada's
Python package index.

### Tagging releases

Releases must be tagged using a format conforming to PEP440, or the packaging
will fail.  Typical deployment releases are such as:

* v0.5

Note the `v` is permitted, though not recommended by PEP440.  It has been
included for clarity and should be retained for consistency.

Development or test releases are also allowable under this scheme:

* v0.6rc1
* v0.6.dev23

Deployment releases must be annotated with a description of the major updates
encapsulated in this release, and this description should be mirrored in [the
changelog](CHANGELOG.md).

### Installing from GitLab package registry

Not recommended, because it's ugly and the registry is different for every
project, and you need to create a token--which is left as an exercise (and
punishment) for the reader.  But something like this:

```
$ TOKEN=__token__ PWD=XXX \
  python3 -m pip install --extra-index-url \
  https://${TOKEN}:${PWD}@git.computecanada.ca/api/v4/projects/771/packages/pypi/simple \
  druid-detector
```

The benefit of this is that every build is uploaded to this registry.
