# vi: set softtabstop=2 ts=2 sw=2 expandtab:
# pylint:

import logging
import socket
from pathlib import Path
import datetime
import subprocess
import viewclust as vc
import pandas as pd
import numpy as np
import plotly.graph_objects as go
from paramiko import SSHClient
from scp import SCPClient
from app.detector import Detector
import app

# Prefix for a folder to write figures to for the interim
root_safe_folder = '/data/analysis/'
# Define how many accounts should be looked at in a debugging pass
debug_num = 5

# Accepts a job frame based on the CLI jobs option.
# For now, we will assume this is the only option.
def detect(frame_path, allow_rac, job_type, cluster_prefix, debug=False):
  # Load job info source
  original_frame = pd.read_pickle(frame_path)
  original_frame['timelimit'] = pd.to_timedelta(original_frame['timelimit'],unit='S')
  filter_frame = original_frame.copy()

  # Filter down to def and cpu accounts who have pending
  print(allow_rac)
  if not allow_rac:
    filter_frame = filter_frame[filter_frame['account'].str.contains('def-')]
  filter_frame = filter_frame[filter_frame['account'].str.contains(f'_{job_type}')]
  filter_frame = filter_frame[filter_frame['state'].str.contains('PENDING')]
  accts = filter_frame.account.unique()
  assumed_target = 50
  use_unit = f'{job_type}-eqv'

  # Timestamp information; 30 day is a constant based on Tyler's job polling
  current_date = datetime.datetime.now()
  start_date = current_date - datetime.timedelta(30)
  d_from = start_date.strftime('%Y-%m-%d')
  d_to = current_date.strftime('%Y-%m-%dT%H:%M:%S')

  # Use timestamp to make safe place to save figures
  safe_folder = root_safe_folder + cluster_prefix + '/' + d_to + '/'
  Path(safe_folder).mkdir(parents=True, exist_ok=True)

  # Accumulator vars
  rel_queue_h=[]
  c_delta=[]
  chq_load=[]
  firstjob = []
  lastjob = []
  jobs_total = []
  submit_accum = []

  # Loop accross the filtered accounts
  acct_count = 0
  for acct in accts:
    logging.info('Working on account: %s', acct)
    acct_count = acct_count + 1
    # Don't do all of the accounts if we're debugging
    if debug and acct_count >= debug_num:
      continue

    # Grab jobs based on original frame
    job_frame = original_frame[original_frame['account'].str.contains(acct)]

    # TODO(James) This try/except block is problematic because it will ignoge any account
    # that causes a job_use failure... this includes accounts that have pending jobs
    # but no completed jobs. Instead job_use should be made to be more robust to
    # account job record states
    try:
      target, queued, running, _ = vc.job_use(job_frame, d_from,
                                                  assumed_target, d_to=d_to,
                                                  use_unit=use_unit)
    except AttributeError:
      rel_queue_h.append(None)
      c_delta.append(None)
      chq_load.append(None)
      firstjob.append(None)
      lastjob.append(None)
      jobs_total.append(None)
      submit_accum.append(None)
      print('No jobs left in the queue for this account')
      continue

    # Usage of running jobs
    _, _, run_running, _ = vc.job_use(job_frame, d_from, target, d_to=d_to,
                                      use_unit=use_unit, job_state='running', time_ref='req')

    # Waiting load of queued jobs
    _, _, q_queued, _ = vc.job_use(job_frame, d_from, target, d_to=d_to,
                                   use_unit=use_unit, job_state='queued', time_ref='horizon+req')

    # Imediate usage time series of completed jobs
    _, _, submit_run, _ = vc.job_use(job_frame, d_from, target, d_to=d_to,
                                     use_unit=use_unit, job_state = 'complete', time_ref='sub')

    # Imediate requested time series of completed jobs
    _, _, submit_req, _ = vc.job_use(job_frame, d_from, target, d_to=d_to,
                                     use_unit=use_unit, job_state = 'complete', time_ref='sub+req')


    # Generate user figures and save in known location by time/user
    figure_folder = safe_folder + acct
    Path(figure_folder).mkdir(parents=True, exist_ok=True)
    # acct_short = acct.split('_')[0]
    _ = vc.insta_plot(target, queued, running,
            fig_out=figure_folder + '/' + acct +'_'+'insta_plot.html',
            submit_run=submit_run,
            submit_req=submit_req,
            running=run_running,
            queued=q_queued,
            query_bounds=True)
    _ = vc.cumu_plot(target, queued, running,
            fig_out=figure_folder + '/' + acct + '_' + 'cumu_plot.html',
            query_bounds=False)

    # Extra append info
    firstjob.append(job_frame['jobid'].min())
    lastjob.append(job_frame['jobid'].max())
    jobs_total.append(len(job_frame.index))
    submit_accum.append(list(job_frame['user'].unique()))

    # Calculate the cumulative target and usage over the query period
    run_cumu = np.cumsum(running).divide(len(running))
    cumu_rel_delta=run_cumu[-1]/50
    cumu_rel_delta=1-cumu_rel_delta
    cumu_rel_delta = max(cumu_rel_delta, 0)
    c_delta.append(cumu_rel_delta)

    # get the horizon queue measure as the mean of pending queued resources beyond the ref_time.
    horizon_q=q_queued[pd.to_datetime(d_to):]
    horizon_q_mean=horizon_q.mean()

    horizon_target_mean = 50

    chq_load.append((horizon_q_mean/horizon_target_mean)*cumu_rel_delta)

    # Calculate the relative queue horizon as the sum horizon queue sum divided by the horizon target sum.
    rel_queue_h.append(horizon_q_mean/horizon_target_mean)

  # Dump list to series
  accts_series = pd.Series(accts)
  rel_queue_series = pd.Series(rel_queue_h)
  c_delta_series = pd.Series(c_delta)
  chq_load_series = pd.Series(chq_load)

  # Series to frame
  frame = { 'account': accts_series,
            'rel_queue_h': rel_queue_series,
            'cumu_delta': c_delta_series,
            'chq_load': chq_load_series,
            'firstjob': pd.Series(firstjob),
            'lastjob': pd.Series(lastjob),
            'submitters': pd.Series(submit_accum),
            'jobs_total': pd.Series(jobs_total)}

  result = pd.DataFrame(frame)
  result=result.sort_values(by=['chq_load','rel_queue_h'], ascending=False)

  # Summary figure
  fig_vals = go.Figure()
  fig_vals.add_trace(go.Scatter(x=result['account'], y=result['rel_queue_h'],
                     mode='lines+markers',
                     name='pending',
                     marker_color='rgba(160,160,160, .8)'))
  fig_vals.add_trace(go.Scatter(x=result['account'], y=result['cumu_delta'],
                     mode='lines+markers',
                     name='use',
                     marker_color='rgba(80,80,220, .8)'))
  fig_vals.add_trace(go.Scatter(x=result['account'], y=result['chq_load'],
                     mode='lines+markers',
                     name='use',
                     marker_color='rgba(220,60,60, .8)'))
  fig_vals.write_html(safe_folder+'burst_rank.html')

  # Format dict object for manager handoff
  report_handoff = []
  for _, account_info in result.iterrows():
    if not account_info.isnull().values.any():
      report_dict = {}
      split_sched_account = account_info['account'].split('_')
      report_dict['account'] = split_sched_account[0]
      report_dict['resource'] = split_sched_account[1]
      report_dict['pain'] = account_info['chq_load']
      report_dict['firstjob'] = account_info['firstjob']
      report_dict['lastjob'] = account_info['lastjob']
      report_dict['submitters'] = account_info['submitters']
      report_summary = {}
      report_summary['num_jobs'] = account_info['jobs_total']
      report_summary['old_pain'] = account_info['rel_queue_h']

      report_dict['summary'] = report_summary
      report_handoff.append(report_dict)

  logging.info("Attempting to zip figure archive...") # Below is a bit ugly, but safe.
  zip_file = safe_folder.rstrip('/') + '.zip'
  subprocess.run(['zip', '-r', zip_file, safe_folder], check=True)

  logging.info("Finished burst report:\n(%s)", report_handoff)
  return zip_file, report_handoff

# ---------------------------------------------------------------------------
#                                                           Bursts Detector
# ---------------------------------------------------------------------------

class Bursts(Detector):

  _name = 'bursts'

  @classmethod
  def define_configuration(cls, config):
    """
    Given a MergeConfSection object, define configuration items for the
    application configuration which will be later merged and made available.

    Arguments:
      config: A MergeConfSection object defined for this Detector class.
    """
    super().define_configuration(config)

    config.add('cluster', mandatory=True)
    config.add('allow_rac', mandatory=True)
    config.add('job_type', mandatory=True)
    config.add('ssh_key', mandatory=True)
    config.add('download_host', mandatory=True)
    config.add('download_retries', value=3, type=int)
    config.add('upload_retries', value=3, type=int)
    config.add('jobs', cli=True, description='Specify an explicit job frame to load (for testing)')
    config.add('many', cli=True, description='Specify how many accounts to examine (for testing)')
    #config.add('RETRIES_UPLOAD', 1, True)
    #config.add('RETRIES_DOWNLOAD', 1, True)
    #config.add('RETRIES_BURST_UPLOAD', 1, True)
    #config.add('RETRIES_BURST_UPLOAD_DELAY', 1, True)

  @classmethod
  def detect(cls):
    """
    Detect the things and report the things.

    This method handles detection of reportable cases and compiles a list to
    contribute to the overall report.
    ```
      [
        {                     # this is a list of dicts describing each case.
          'account': ...,     # here is a datum reported for all case types
          'resource': ...,    # here is a datum specific to this type of case
          'summary': ...,     # all cases support a summary, though optional
          ...                 # there are definitely others
        },
        ...                   # this list can be one, many, or zero length
      ],
    ```
    """
    # TODO: is this redundant over host_dl?
    #cluster = cls._config.cluster
    allow_rac = cls._config.allow_rac
    job_type = cls._config.job_type
    scp_key = cls._config.ssh_key
    host_dl = cls._config.download_host
    upload_retries = cls._config.upload_retries
    download_retries = cls._config.download_retries
    job_frame = cls._config.jobs
    if cls._config.many is not None:
      # pylint: disable=global-statement
      global debug_num
      debug_num = cls._config.many

    cluster_prefix = host_dl[:3]

    # Do a bunch of detectorating
    if job_frame is not None: # CLI
      zip_file, bursts = detect(job_frame, allow_rac, job_type, cluster_prefix, debug=app.config.debug)
    else: # SCP from pseudo user
      attempts = 0
      download_complete = False
      while attempts < download_retries:
        try:
          logging.info('Attempting dataframe download...')
          ssh = SSHClient()
          ssh.load_system_host_keys()
          ssh.connect(host_dl, username='beam', key_filename=scp_key, gss_deleg_creds=False)
          with SCPClient(ssh.get_transport()) as scp:
            scp.get(f'/project/def-tk11br/{cluster_prefix}_all.pkl', local_path=root_safe_folder)
            download_complete = True
        except socket.gaierror as err:
          attempts += 1
          logging.info('Download failed: %s', err)

        if download_complete:
          logging.info('Download success')
          break

      zip_file, bursts = detect(root_safe_folder+f'/{cluster_prefix}_all.pkl',
        allow_rac, job_type, cluster_prefix, debug=app.config.debug)

    # Transfer zip here before burst data is reported.  Should the Detector
    # class offer a separate method so this upload can be carried out
    # independently or in a different order?
    attempts = 0
    logging.info('Transferring zip file %s', zip_file)
    while attempts < upload_retries:
      rc = subprocess.call(['./generic_upload.sh', zip_file])
      if rc == 0:
        break
      attempts += 1
      logging.error('Transfer failed with rc %d (attempt %d of %d)',
        rc, attempts, upload_retries)

    return bursts
