# vi: set softtabstop=2 ts=2 sw=2 expandtab:
# pylint: disable=unspecified-encoding

import sys
import subprocess
import time
import datetime
from app.detector import Detector

# ---------------------------------------------------------------------------
#                                                         Old Jobs Detector
# ---------------------------------------------------------------------------

# Function to call the squeue command to return a list of jobs sorted in descending order by job age
def querySlurm(verbose = False, attempts = 5):
  try:
    # squeue --Format="JobID:. ,Account:. ,UserName:. ,SubmitTime:. ,tres-alloc:. " -S "V"
    # Command below is a little broken, does not format the first requested field correctly.
    # Cause is currently unknown. Recommend wrapping in shell script to drop output to a text file
    # Import using readFile() instead.
    queueQuery = subprocess.check_output(["/opt/software/slurm/bin/squeue",
                                          '--Format=Account:. ,JobID:. ,UserName:. ,SubmitTime:. ,tres-alloc:. ',
                                          '-S', 'V'])
    return queueQuery
  except subprocess.CalledProcessError:
    if attempts > 0:
      time.sleep(10)
      return querySlurm(verbose, attempts-1)
    return False

# Function to read in a text file simulating a slurm queue (or a cached slurm queue)
def readFile(filePath):
  output = []
  try:
    with open(filePath, mode='r') as queueQuery:
      tempList = list(queueQuery)
      for line in tempList:
        line = line.replace("\n","")
        output.append(line)
  except FileNotFoundError:
    print(filePath +  " could not be opened. Please ensure that the file exists and is accessible.")
    sys.exit(10)
  return output

def processHeader(queueQuery):
  # Return the first line of the queueQuery, split along sapce character
  return queueQuery[0].split(" ")

# Create a nested dict containing jobs indexed on JOBID
# (This will work over the timeframe that we are considering, duplicate JOBIDs are extremely unlikely.)
def processQueue(queueQuery):
  jobs = {}
  header = processHeader(queueQuery)
  jobIDPos = header.index("JOBID")

  # Remove the first line, since it contains the headers
  queueQuery.pop(0)
  for line in queueQuery:
    # Split up each line along |
    elements = line.split(" ")
    jobs[str(elements[jobIDPos])] = {}
    for i, element in enumerate(elements):
      jobs[str(elements[jobIDPos])][header[i]] = element
  return jobs

# Listifies tresAll, returns as JSON
def processTresAlloc(tres):
  output = {}
  for each in tres.split(","):
    output[each.split("=")[0]] = each.split("=")[-1]
  return output

# Function to process data and transform jobs into expected output
def processOutput(jobs, min_age):
  oldjobs = []
  # Track accounts and how many times they have a job in queue
  oldAccounts = {}
  for each in jobs:
    age = round(float((datetime.datetime.now() - datetime.datetime.strptime(jobs[each]["SUBMIT_TIME"],
                                                                  "%Y-%m-%dT%H:%M:%S")).total_seconds())/(60*60*24))
    # Age Filter
    if age > int(min_age):
      replace = False
      jobid = jobs[each]["JOBID"]
      account = jobs[each]["ACCOUNT"].split("_")[0]
      user = jobs[each]["USER"]
      resource = jobs[each]["ACCOUNT"].split("_")[-1].upper()
      tresAlloc = processTresAlloc(jobs[each]["TRES_ALLOC"])
      if account not in oldAccounts:
        count = 1
        replace = True
      elif age > oldAccounts[account]["age"]:
        count = oldAccounts[account]["summary"]["count"]
        count += 1
        replace = True
      else:
        oldAccounts[account]["summary"]["count"] += 1
      if replace:
        oldAccounts[account] = {
          "job_id": jobid,
          "account": account,
          "submitter": user,
          "resource": resource,
          "age": age,
          "summary": {
            "count": count,
          },
        }
        for key, alloc in tresAlloc.items():
          oldAccounts[account]["summary"][key] = alloc

  for account in oldAccounts.values():
    oldjobs.append(account)
  return oldjobs

class OldJobs(Detector):

  _name = 'oldjobs'

  @classmethod
  def define_configuration(cls, config):
    """
    Given a MergeConfSection object, define configuration items for the
    application configuration which will be later merged and made available.

    Arguments:
      config: A MergeConfSection object defined for this Detector class.
    """
    super().define_configuration(config)
    #config.add('threshold', type=str, value='30d')
    config.add('min_age', type=int, mandatory=True)
    config.add('file_path', type=str, mandatory=True)

  @classmethod
  def detect(cls):
    """
    Detect the things and report the things.

    This method handles detection of reportable cases and compiles a list to
    contribute to the overall report.
    ```
      [
        {                     # this is a list of dicts describing each case.
          'account': ...,     # here is a datum reported for all case types
          'resource': ...,    # here is a datum specific to this type of case
          'summary': ...,     # all cases support a summary, though optional
          ...                 # there are definitely others
        },
        ...                   # this list can be one, many, or zero length
      ],
    ```
    """
    min_age = cls._config.min_age
    file_path = cls._config.file_path

    # Read in file
    queue = readFile(file_path)
    # Or query Slurm queue
    # queue = querySlurm()
    # Process file
    jobs = processQueue(queue)
    output = processOutput(jobs, min_age)
    return output
