"""
The signedapi Python module provides a basic client library for making signed
API requests.
"""

from email.utils import formatdate
import hmac
import base64
import requests

class SignedApi:
  """
  Class for making signed requests to a compliant API server.
  """

  def __init__(self, tag, uri, access, secret):
    """
    Initializes SignedAPI class for use with a specific URI.
    
    Args:
      tag (str): Short string identifying application.  Must match server.
      uri (str): Base URI for API server.  Example:
        'https://service.example.org/api'
      access (str): API key identifier
      secret (str): API key
    """

    self._tag = tag
    self._uri = uri
    self._access = access
    self._secret = secret.encode()

  def _make_headers(self, verb, resource):
    """
    Helper method to generate headers based on target resource and HTTP verb.

    Args:
      verb (str): HTTP verb (see
        https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods)
      resource (str): Resource path, not including what was in the base URI.
    """

    # need current timestamp in RFC2822 format.
    timestamp = formatdate(localtime=True)

    # build request string to digest
    digestible = "{} {}\n{}".format(verb, resource, timestamp)

    # create digest
    h = hmac.new(self._secret, digestmod='sha256')
    h.update(digestible.encode())
    digest = base64.b64encode(h.digest()).decode('utf-8')

    # headers to add to request
    headers = {
      'Date': timestamp,
      'Authorization': "{} {} {}".format(self._tag, self._access, digest)
    }

    return headers

  def get(self, resource):
    """
    Retrieve the specified resource or a collection of like resources.

    Args:
      resource (str): Resource path, not including what was in the base URI.

    Returns:
      Resource definition, for a single resource requested; or a collection of
      resource descriptions, for multiple resources.
    """

    uri = self._uri + resource
    headers = self._make_headers("GET", resource)
    r = requests.get(uri, headers=headers)
    return r

  def post(self, resource, data):
    """
    Submit a new resource.

    Args:
      resource (str): Resource path, not including what was in the base URI.
      data (map): Definition of the resource.
    """

    uri = self._uri + resource
    headers = self._make_headers("POST", resource)
    r = requests.post(uri, json=data, headers=headers)
    return r
